import argparse
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import json
import math
import networkx
import pandas as pd
from scipy.stats import rankdata
import sys
import time

# example: python info_coverage.py -i Time-4b-coverage.csv.xz -c changeset.json -b 4
# --> 'org.joda.time.Partial:464'

def arg_parser():
    parser = argparse.ArgumentParser(description = 'Info')
    parser.add_argument('-i', '--input', help='input .xz file')
    parser.add_argument('-c', '--change-set', required=True, help='change set (json file)')
    parser.add_argument('-b', '--bugID', required=True, help='defects4j bugID')
    parser.add_argument('-m', '--method', choices=['average', 'min', 'max'], help='method of the rank calc.')
    parser.add_argument('-a', '--algorithms', nargs="+", choices=get_algorithms(), default=get_algorithms())

    param_dict = {}
    args = parser.parse_args()
    param_dict["input"] = args.input
    param_dict["change-set"] = args.change_set
    param_dict["bugID"] = args.bugID
    param_dict["method"] = "average" if args.method is None else args.method
    param_dict["algorithms"] = args.algorithms
    return param_dict


def get_algorithms():
    return ['Barinel', 'DStar', 'Jaccard', 'Ochiai', 'SorensenDice', 'Tarantula']


def get_tests(matrix):
    return matrix['name'].tolist()


def get_failed_tests(matrix):
    return matrix[matrix['outcome'] == 0]['name'].tolist()


def get_passed_tests(matrix):
    return matrix[matrix['outcome'] == 1]['name'].tolist()


def get_statements(matrix):
    st = matrix.columns.tolist()
    st.remove('name')
    st.remove('outcome')
    return st


def get_graph(matrix, statements):
    graph = networkx.Graph()
    edges = list()
    for st in statements:
        edges = edges + [(st, t) for t in matrix[matrix[st] == 1]['name'].tolist()]
    graph.add_edges_from(edges)
    return graph


def get_spectra(graph, statements, tests, failed_tests):
    spectra = {}
    for st in statements:
        spectra[st] = {}
        if st not in list(graph.nodes()):
            spectra[st]['ef'] = 0
            spectra[st]['ep'] = 0
            spectra[st]['nf'] = 0
            spectra[st]['np'] = 0
        else:
            covered_tests = set(list(graph.neighbors(st)))
            spectra[st]['ef'] = len(covered_tests.intersection(set(failed_tests)))
            spectra[st]['ep'] = len(covered_tests) - spectra[st]['ef']
            number_of_not_covered_tests = len(tests) - len(covered_tests)
            spectra[st]['nf'] = len(failed_tests) - spectra[st]['ef']
            spectra[st]['np'] = number_of_not_covered_tests - spectra[st]['nf']
    return spectra


def get_scores(statements, spectra, algorithms):

    def barinel(ef, ep):
        return float(ef)/float(ef+ep) if ef+ep != 0 else 0.0

    def dstar(ef, ep, nf):
        return float(ef*ef)/float(ep+nf) if ep+nf != 0 else 0.0

    def jaccard(ef,ep, nf):
        return float(ef)/float(ef+nf+ep) if ef+nf+ep != 0 else 0.0

    def ochiai(ef, ep, nf):
        return float(ef)/float(math.sqrt((ef+nf)*(ef+ep))) if ef+nf > 0 and ef+ep > 0 else 0.0

    def sorensenDice(ef, ep, nf):
        return float(2*ef)/float(2*ef + nf + ep) if 2*ef + nf + ep > 0 else 0.0

    def tartantula(ef, ep, nf, np):
        a = float(ef)/float(ef+nf) if ef+nf > 0 else 0
        b = float(ef)/float(ef+nf) if ef+nf > 0 else 0
        c = float(ep)/float(ep+np) if ep+np > 0 else 0
        if b+c > 0.0:
            return float(a)/float(b+c)
        return 0.0

    def calc(algorithm, ef, ep, nf, np):
        if algorithm == 'Barinel':
            return barinel(ef, ep)
        elif algorithm == 'DStar':
            return dstar(ef, ep, nf)
        elif algorithm == 'Jaccard':
            return jaccard(ef, ep, nf)
        elif algorithm == 'Ochiai':
            return ochiai(ef, ep, nf)
        elif algorithm == 'SorensenDice':
            return sorensenDice(ef, ep, nf)
        elif algorithm == 'Tarantula':
            return tartantula(ef, ep, nf, np)
        else:
            pass

    scores = {}
    for alg in algorithms:
        scores[alg] = {}
        for st in statements:
            scores[alg][st] = calc(alg, spectra[st]['ef'], spectra[st]['ep'], spectra[st]['nf'], spectra[st]['np'])
    return scores


def get_ranks(scores, algorithms, method):
    def calc(results, method):
        return dict(zip(results.keys(), rankdata([-i for i in results.values()], method=method)))

    ranks = {}
    for alg in algorithms:
        ranks[alg] = calc(scores[alg], method)
    return ranks


def get_buggy_statements(changeset, bugID):
    f = open(changeset, "r")
    return set(json.load(f)[str(bugID)])


def get_info(graph, spectra, algorithms, scores, ranks, statements, buggy_statements, tests, failed_tests):
    print("The name of the test or statement elements (exit -> \'Exit\') :")
    for entity in sys.stdin:
        entity = entity.rstrip()
        if 'Exit' == entity:
            break
        else:
            if entity in statements:
                try:
                    print(entity)
                    print("\tis buggy? - ", "yes" if entity in buggy_statements else "no")
                    print("\tef - ", str(spectra[entity]["ef"]))
                    print("\tep - ", str(spectra[entity]["ep"]))
                    print("\tnf - ", str(spectra[entity]["nf"]))
                    print("\tnp - ", str(spectra[entity]["np"]))
                    print("\tscore:")
                    for alg in algorithms:
                        print("\t\t", alg, " - ", scores[alg][entity] if entity in scores[alg] else " -- ")
                    print("\trank:")
                    for alg in algorithms:
                        print("\t\t", alg, " - ", ranks[alg][entity] if entity in ranks[alg] else " -- ")
                    print("\tcovered by tests:")
                    if entity in list(graph.nodes()):
                        for t in list(graph.neighbors(entity)):
                            print("\t\t", t)
                except:
                    pass
            elif entity in tests:
                try:
                    print(entity)
                    print("\tis failed test? - ", "yes" if entity in failed_tests else "no")
                    print("\tcovered statements:")
                    if entity in list(graph.nodes()):
                        for t in list(graph.neighbors(entity)):
                            print("\t\t", t, " - ", "buggy" if t in buggy_statements else "not buggy")
                except:
                    pass
            else:
                print("This entity does not exist")



param_dict = arg_parser()
start_time = time.time()
print("Start: %s" % datetime.now().strftime("%d/%m/%Y %H:%M:%S"))

print("Read the (coverage) matrix")
matrix = pd.read_csv(param_dict['input'])

print("Read the buggy statements")
buggy_statements = get_buggy_statements(param_dict['change-set'], param_dict['bugID'])

print("Read the test- and the statement-names")
tests = get_tests(matrix)
failed_tests = get_failed_tests(matrix)
statements = get_statements(matrix)

print("Convert (matrix) to graph")
graph = get_graph(matrix, statements)

print("Calculate the spectra")
spectra = get_spectra(graph, statements, tests, failed_tests)

print("Calculate the score")
scores = get_scores(statements, spectra, param_dict['algorithms'])

print("Calculate the ranks")
ranks = get_ranks(scores, param_dict['algorithms'], param_dict['method'])

print("End: %s" % datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
print("( %s sec. )" % (time.time() - start_time))

get_info(graph, spectra, param_dict['algorithms'], scores, ranks, statements, buggy_statements, tests, failed_tests)